#local includes

#external includes
import re
import logging
import importlib.util
import inspect
import traceback
import yaml

def safeGetMoudle(rawModule):
    justModule = rawModule.split("/")[-1]
    return justModule.replace(".py", "")

def getConfigDictFromFile(path):
    if path is not None:
        with open(path) as file:
            return yaml.load(file, yaml.FullLoader)
    return None

def safeParse(rawValue, settings = None):
    #much safe
    logging.warning("safe parse called for this value")
    logging.warning("returning raw value")
    return rawValue

def stringParse(rawValue, settings = None):
    return rawValue

def valueParse(rawValue, settings=None):
    #need to change this in someway
    return rawValue.strip()

def listParse(rawValue, settings=None):
    reDelim = ", "
    if settings is not None and 'delimiter' in settings:
        if type(settings['delimiter']) is list and len(settings['delimiter'])>0:
            reDelim = ""
            for delim in settings['delimiter']:
                reDelim += str(delim)
                reDelim += "|"
            reDelim = reDelim[:-1]
    tempList = [item.strip() for item in re.split(reDelim, rawValue)]
    return list(filter(None, tempList))

def findParseFunc(valueType):
    if valueType == "list":
        return listParse
    elif valueType == "value":
        return valueParse
    elif valueType == "string":
        return stringParse
    else:
        return safeParse

class parser:
    def __init__(self, settings=None, settingsPath=None):
        self.m_settings = settings if settings is not None else getConfigDictFromFile(settingsPath)
        if self.m_settings is not None and 'identifierValueType' in self.m_settings:
            self.m_valueTypeDict = self.m_settings['identifierValueType']

    def parseRawValue(self, identifier, rawValue):
        vTSafe = str(self.m_valueTypeDict.get(identifier, "safe")).lower()
        parseFunc = findParseFunc(vTSafe)
        return parseFunc(rawValue, settings=self.m_settings.get(vTSafe,None))

    def loadFile(self, filePath, marker):
        objSpecs = {}
        try:
            spec = importlib.util.spec_from_file_location(filePath, filePath)
            mod = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(mod)
        except Exception:
            logging.error(traceback.format_exc())
        else:
            module = safeGetMoudle(filePath)
            objSpecs[module] = {}
            for funcTupl in inspect.getmembers(mod, inspect.isfunction):
                specDict = {}
                try:
                    doc = inspect.getdoc(funcTupl[1])
                except Exception:
                    logging.error(traceback.format_exc())
                else:
                    if doc is not None:
                        for line in doc.split("\n"):
                            ident, val = self.handleMarkerLine(line, marker)
                    
                            if ident is not None and val is not None:
                                if ident not in specDict:
                                    specDict[ident] = val

                    specDict['args'] = inspect.getargspec(funcTupl[1])[0]
                    specDict['call'] = funcTupl[1]
                    objSpecs[module][funcTupl[0]] = specDict
        return objSpecs

    def handleMarkerLine(self, line, marker):
        index = line.find(marker)   #find where marker is if any
        identifier = None
        value = None
        if index > -1 and index < len(line) - 1:    #if marker in line and not end of line
            split = line[index + len(marker) :].split(maxsplit=1) #split line after marker into identifer and after
            identifier = split[0].replace(":","") #remove colon if there is
            value = self.parseRawValue(identifier, split[1]) #use parser to handle raw val

        return identifier, value
        


#local includes
from . import parser

#external includes
import collections

class libObject:
    def __init__(self, obj, details = None):
        self.m_obj = obj
        self.m_module = None
        self.m_name = None
        self.m_details = {}
        self.m_reqs = {}
        if self.m_obj is not None:
            if obj.__module__ is not None:
                self.m_module = parser.safeGetMoudle(obj.__module__)
            self.m_name = obj.__name__

        if isinstance(details, collections.Mapping):
            self.m_details = details


#local includes
from . import parser
from . import libObject

#external includes
import logging
import inspect

#requiredFunc must be first
requiredIdentifiers = ['requiredFunc', 'requiredName', 'requiredCategory']

def reqSetDiscard(objToCheck, libObjDiscard):
    if 'requiredFunc' in objToCheck.m_reqs:
        objToCheck.m_reqs['requiredFunc'].discard(libObjDiscard.m_name)

    #make sure that requyiredFunc is first
    for identifier in requiredIdentifiers[1:]:
        noRequired = identifier.replace("required","").lower()
        if identifier in objToCheck.m_reqs and noRequired in libObjDiscard.m_details:
            objToCheck.m_reqs[identifier].discard(libObjDiscard.m_details[noRequired])

class library():
    def __init__(self, parser):
        self.m_objSpecs = {}
        self.m_parser = parser

    def loadFile(self, filePath, marker):
        #add moduleDict to current spec dict
        self.m_objSpecs = {**self.m_objSpecs, **self.m_parser.loadFile(filePath, marker)}
        
    def printObjSpecs(self):
        print(self.m_objSpecs)

    def prettyPrintObjSpecs(self, specArgList = []):
        if type(specArgList) is list and len(specArgList) > 0:
            for module in self.m_objSpecs:
                print(module + ":")
                for func in self.m_objSpecs[module]:
                    print(func + ":")
                    for objSpec in self.m_objSpecs[module][func]:
                        for specArg in specArgList:
                            if specArg in objSpec:
                                print(str(specArg) + ": " + str(objSpec[str(specArg)]))
        else:
            for module in self.m_objSpecs:
                print(module + ":")
                for func in self.m_objSpecs[module]:
                    print(func + ":")
                    if 'call' in self.m_objSpecs[module][func]:   
                        print(str(self.m_objSpecs[module][func]['call']))
                    for key, value in self.m_objSpecs[module][func].items():
                        if key != 'call':
                            print("\t\t" + str(key) + ": " + str(value))


    def reqSort(self, libObjs):
        """
            sorts and filters libObjs based on requirements
            items that don't fit in, either don't have all their requirements met 
            or aren't in specs, are filtered out
        """
        #first build specific requirements
        for libObj in libObjs:
            if libObj is None or libObj.m_module is None or libObj.m_name is None or libObj.m_name == "":
                logging.warning("error finding object during sort")
                logging.warning("obj: " + str(libObj.m_obj))
            else:
                if libObj.m_module in self.m_objSpecs and libObj.m_name in self.m_objSpecs[libObj.m_module]:
                    objSpec = self.m_objSpecs[libObj.m_module][libObj.m_name]
                    for identifier in requiredIdentifiers:
                        if identifier in objSpec:
                            if identifier not in libObj.m_reqs:
                                libObj.m_reqs[identifier] = set()
                            req = objSpec[identifier]
                            if isinstance(req, list):
                                libObj.m_reqs[identifier].update(req)
                            elif isinstance(req, str):
                                libObj.m_reqs[identifier].add(req)
                                

        newLibObjList = []
        #using the specificly built object requirements, build the sorted and filtered list
        change = True
        while change:
            change = False
            for libObj in libObjs[:]:   #if this loop ends with never going inside the if then it will break
                if all(not reqSet for reqSet in libObj.m_reqs.values()):  #check if requirement set is empty
                    change = True
                    newLibObjList.append(libObj)
                    for obj in libObjs:
                        #remove that item from all requirement sets
                        reqSetDiscard(obj, libObj)
                    libObjs.remove(libObj)
                    break  #break here to favor that libObjs were put in, earlier items will always be added if reqs met before later items
 

        return newLibObjList


def testFunc(x):
    """
    @requiredFunc: otherFileFunc
    """
    return x + 1

def moreFunc(y, z):
    """
    @description: basic multiplication function
    """
    val = y * z
    return val

def noDocFunc(test):
    testString = str(test) + " this is a test"
    return testString

def myNameIsTestFunc2():
    """
    @requiredFunc: moreFunc
    @requiredName: Close
    """
    x = 2


def otherFunc():
    """
    @description: i do literally nothing like the rest
    """
    return "yo"


def ema(x):
    """
    @name: ema
    @requiredFunc: testFunc
    """
    return x + 1

def emaAverage(y, z):
    """
    @name: emaAverage
    @requiredFunc ema
    @description: basic ema Average function
    """
    val = y * z
    return val

def emaAverageAverage(test):
    """
    @name: emaAvgAvg
    @requiredFunc emaAverage
    @description: basic ema Average Average function
    """
    testString = str(test) + " this is a test"
    return testString

def otherFileFunc():
    """

    """
    return 42


#local includes
from . import library
from .libObject import libObject

#external includes
import logging

"""
Group of lib objects, correspond to a library
"""

class libGroup():
    def __init__(self, library=None):
        self.m_library = library
        self.m_sorted = True
        self.m_objs = []

    def insert(self, obj):
        if obj is not None:
            libObj = None
            if not isinstance(obj, libObject):
                libObj = libObject(obj)
            else:
                libObj = obj
                
            if libObj.m_module is None or libObj.m_name is None:
                logging.warning("error creating libObj from obj:")
                logging.warning(obj)
    
            self.m_objs.append(libObj)
            self.m_sorted = False


    def sort(self):
        self.m_objs = self.m_library.reqSort(self.m_objs)
        self.m_sorted = True
        
    def getObjs(self):
        if not self.m_sorted:
            self.sort()
        return self.m_objs

    def changeLibrary(self, library):
        self.m_library = library
        self.sort()

    def prettyPrint(self):
        for obj in self.m_objs:
            print(str(obj.m_module) + "." + str(obj.m_name))


#!/usr/bin/python3
import setuptools
import io
import os
import sys

# Package meta-data
NAME = 'requiremental'
DESCRIPTION = 'A requirement managemet tool.'
URL = 'https://github.com/rmallow/requiremental'
EMAIL = 'mail@rmallow.com'
AUTHOR = 'Robert Mallow'
REQUIRES_PYTHON = '>=3.6.0'
VERSION = None

# Packages required
REQUIRED = [
		'PyYaml'
		]

here = os.path.abspath(os.path.dirname(__file__))

with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
	long_description = '\n' + f.read()

# Load the package's __version__.py
about = {}
if not VERSION:
	with open(os.path.join(here, NAME, '__version__.py')) as f:
		exec(f.read(), about)
else:
	about['__version__'] = VERSION

setuptools.setup(
	name=NAME,
	version=about['__version__'],
	description=DESCRIPTION,
	long_description=long_description,
	long_description_content_type="text/markdown",
	author=AUTHOR,
	author_email=EMAIL,
	python_requires=REQUIRES_PYTHON,
	url=URL,
	packages=setuptools.find_packages(),
	install_requires=REQUIRED,
	license='MIT',
	classifiers=[
		'License :: OSI Approved :: MIT License',
		'Programming Language :: Python',
		'Programming Language :: Python :: 3',
		'Programming Language :: Python :: 3.6',
		'Programming Language :: Python :: 3.7',
		'Programming Language :: Python :: 3.8',
	],
)